import typing
from . import types, utils, models


class prepr(types.prepr):
    def __init__(self, inst, variable_name: str = None,
                 note: str = None) -> None:
        """Create a pretty-formatted string for class representations.
        
        Parameters
        ----------
        inst : any
            The class instance being represented.
        variable_name : str, optional
            The name of the variable that the representation will be assigned to
            when printed.
        note : str, optional
            A comment about the class.
        
        Example usage
        -------------
        ```
        class example_class:
            def __init__(self, a, b, c = None):
                self.a = a
                self.b = b
                self.c = c
                # a value which will be set
                # somewhere down the line:
                self.something_else = None
            def __repr__(self):
                R = prepr(self)
                R.args(self.a, self.b)
                R.kwarg("c", self.c)
                R.attr(self.something_else)
                return R.build()
        ```

        """
        self._exc = None
        try:
            self._inst = inst
            self._variable: str = models.settings.csh.f_variable(variable_name
                or "__" + type(inst).__name__.lower() + "__")
            self._note: str = models.settings.csh.f_comment(
                models.settings.comment + note if note else "")
            self._name: str = models.settings.csh.f_class(type(inst).__name__)
            self._args: list[str] = []
            self._kwargs: dict[str, str] = {}
            self._attrs: dict[str, str] = {}
        except Exception as exc:
            self._exc = exc
    

    def arg(self, v) -> types.prepr:
        """Add a positional argument.
        
        """
        if not self._exc:
            try:
                self._args.append(v)
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self
    

    def args(self, *v) -> types.prepr:
        """Convenience method for adding multiple arguments in one function
        call.
        
        """
        if not self._exc:
            try:
                self._args.extend(v)
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self
    

    def kwarg(self, k, v, d: typing.Any = types.MISSING) -> types.prepr:
        """Add a keyword argument.

        `d` is optional and can be set to any value. If `v` (the value) is equal
        to `d` (the default value), it will be ommited from the representation.
        
        """
        if not self._exc:
            try:
                if (d != types.MISSING and d == v):
                    return self
                self._kwargs[k] = v
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self
    

    def kwargs(self, d: typing.Any = types.MISSING, **kv) -> types.prepr:
        """Convenience method for adding multiple keyword arguments in one
        function call.

        `d` is optional and can be set to any value. If and value in `kv` is
        equal to `d` (the default value), the pair will be ommited from the
        representation.
        
        """
        if not self._exc:
            try:
                if d != types.MISSING:
                    temp = {}
                    for k, v in kv.items():
                        if v == d:
                            continue
                        temp[k] = v
                else:
                    self._kwargs.update(kv)
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self


    def attr(self, k, v, d: typing.Any = types.MISSING) -> types.prepr:
        """Add an attribute.

        This should only be used for variable attributes that change sometime
        after class initialization.

        `d` is optional and can be set to any value. If `v` (the value) is equal
        to `d` (the default value), it will be ommited from the representation.
        
        """
        if not self._exc:
            try:
                if d != types.MISSING and d == v:
                    return self
                self._attrs[k] = v
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self
    

    def attrs(self, d: typing.Any = types.MISSING, **kv) -> types.prepr:
        """Convenience method for adding multiple attributes in one function
        call.

        `d` is optional and can be set to any value. If and value in `kv` is
        equal to `d`, the pair will be ommited from the representation.
        
        """
        if not self._exc:
            try:
                if d != types.MISSING:
                    temp = {}
                    for k, v in kv.items():
                        if v == d:
                            continue
                        temp[k] = v
                else:
                    self._attrs.update(kv)
                return self
            except Exception as exc:
                self._exc = exc
                return self
        return self
    

    def _format_args(self, exc: dict[int, str], i: str, lb: str) -> str:
        """Format the positional and keyword arguments.
        
        """
        formatted_args = [utils.concat(i, lb, utils.format_value(v, exc, i, lb))
                          for v in self._args]
        formatted_kwargs = [utils.concat(i, lb, 
            models.settings.csh.f_argument(k),
            models.settings.csh.f_operator(models.settings.equals),
            utils.format_value(v, exc, i, lb)) for k, v, in
            self._kwargs.items()]
        return utils.indent(models.settings.csh.f_operator(
            models.settings.comma).join(formatted_args + formatted_kwargs), i
            ) if formatted_args or formatted_kwargs else ""
    

    def _format_attrs(self, exc: dict[int, str], lb: str) -> str:
        """Format the attributes.
        
        """
        formatted_attrs = [utils.concat(
            self._variable,
            models.settings.csh.f_operator("."),
            models.settings.csh.f_attribute(k),
            models.settings.csh.f_operator(models.settings.equals),
            utils.format_value(
                v,
                exc,
                models.settings.indent,
                models.settings.line_break
            )
        ) for k, v in self._attrs.items()]
        return lb.join(formatted_attrs)


    def _build_collapsed(self) -> types.pstr:
        """Return `<name>(...)`.
        
        """
        return (
            utils.concat(
                self._name,
                models.settings.csh.f_bracket("("),
                models.settings.csh.f_operator("..."),
                models.settings.csh.f_bracket(")")
            ),
            self
        )


    def _build_simple(self, exc: dict[int, str], i: str, lb: str) -> str:
        """Create the main representation without any attributes or variable
        name.
        
        """
        args = self._format_args(exc, i, lb)
        return utils.concat(
            self._name,
            models.settings.csh.f_bracket("("),
            self._note if "\n" in models.settings.line_break else "",
            args,
            i if args else "",
            lb if args else "",
            models.settings.csh.f_bracket(")")
        )
    
    
    def build(self, simple: bool = False, collapsed: bool = False,
              return_prepr: bool = False) -> types.pstr | types.prepr:
        """Build the representation.
        
        Arguments
        ---------
        simple : bool, optional, default=False
            If True, the attributes and variable name will be ommitted.
        collapsed : bool, optional, default=False
            If True, only the class name will be included with parentheses and
            an ellipsis (for example, `ExampleClass(...)`). Takes precedence
            over `simple`.
        return_prepr : bool, optional, default=False
            A value used internally to resolve issues with recursion. Should not
            be defined by the user.

        """
        if not self._exc:
            try:
                if return_prepr:
                    return self
                if collapsed is True:
                    return self._build_collapsed()

                # create the simple representation now as it
                # used whether or not `simple` is True or False
                instance_exceptions = {id(self._inst): self._variable}
                simple_repr = self._build_simple(
                    instance_exceptions,
                    models.settings.indent,
                    models.settings.line_break
                )

                # if `simple` is True, the simply return simple_repr
                if simple is True:
                    return models.pstr(simple_repr, self)
                
                # create the line break used to separate the
                # simple_repr and each attribute assignment
                lb = utils.concat(
                    models.settings.csh.f_error(models.settings.semicolon),
                    models.settings.indent,
                    models.settings.line_break
                )

                # return the full repr
                return models.pstr(
                    utils.concat(
                        self._variable,
                        models.settings.csh.f_operator(models.settings.equals),
                        simple_repr,
                        (lb + self._format_attrs(instance_exceptions, lb) if
                         self._attrs else "")
                    ),
                    self
                )
            except Exception as exc:
                self._exc = exc
        return models.pstr("\033[32mPreprBuildFailure\033[33m(\033[31m\"" +
                           str(self._exc) + "\"\033[33m)\033[0m", self)



import typing


class MISSING:
    """A type to indicate no value in `prepr.kwarg`.
    
    """


class pstr(str):
    _prepr: "prepr"
    def __new__(self, text, *args, **kwargs): ...
    def __init__(self, text: str, _prepr: "prepr"): ...


class prepr:
    def __init__(self, inst, variable_name: str = None,
                 note: str = None) -> None: ...
    def arg(self, v) -> "prepr": ...
    def args(self, *v) -> "prepr": ...
    def kwarg(self, k, v, d: typing.Any = MISSING) -> "prepr": ...
    def kwargs(self, d: typing.Any = MISSING, **kv) -> "prepr": ...
    def attr(self, k, v, d: typing.Any = MISSING) -> "prepr": ...
    def attrs(self, d: typing.Any = MISSING, **kv) -> "prepr": ...
    def _format_args(self, exc: dict[int, str], i: str, lb: str) -> str: ...
    def _format_attrs(self, exc: dict[int, str], lb: str) -> str: ...
    def _build_collapsed(self) -> pstr: ...
    def _build_simple(self, exc: dict[int, str], i: str, lb: str) -> str: ...
    def build(self, simple: bool = False, collapsed: bool = False,
              return_prepr: bool = False) -> pstr: ...


class Colorspace:
    c_function: str
    c_class: str
    c_string: str
    c_number: str
    c_other: str
    c_variable: str
    c_attribute: str
    c_argument: str
    c_operator: str
    c_comment: str
    c_bracket: str
    c_error: str
    c_boolean: str
    c_enum: str
    c_reset: str



from . import types, models
import textwrap
import inspect
import enum
import types as builtin_types


def concat(*__text: str) -> str:
    """Concatenate any number of `str` objects together.
    
    """
    return "".join(__text)


def attempt_str(value) -> str:
    """Attempt to run `str(__value)`. If an error is encountered, return "???"
    (formatted as error).

    """
    try:
        return str(value)
    except Exception:
        return concat(
            models.settings.csh.f_operator(".").join(
                [models.settings.csh.f_class(class_name) for class_name
                in type(value).__qualname__.split(".")]),
            models.settings.csh.f_bracket("("),
            models.settings.csh.f_error("!!!"),
            models.settings.csh.f_bracket(")"))


def indent(__str: str, __indent: str) -> str:
    """Increases the indent of each line in `__str` by `__indent`.
    
    """
    return textwrap.indent(__str, __indent)


def format_str(v: str) -> str:
    """Format a string value.
    
    """
    return models.settings.csh.f_string(concat("\"",
        v.replace("\\", "\\\\").replace("\"", "\\\""), "\""))

def format_num(v: int | float) -> str:
    """Format a number (int/float) value.
    
    """
    return models.settings.csh.f_number(attempt_str(v))

def format_bool_none(v: bool | None) -> str:
    """Format boolean or None value.
    
    """
    return models.settings.csh.f_boolean(attempt_str(v))

def format_list(v: list, exc: dict[int, str], i: str, lb: str) -> str:
    """Format a list value.
    
    """
    I = "" if models.settings.force_lists_collapsed else i
    LB = "" if models.settings.force_lists_collapsed else lb
    values = [concat(LB, format_value(V, exc, I, LB)) for V in v]
    return concat(
        models.settings.csh.f_bracket("["),
        indent(
            models.settings.csh.f_operator(models.settings.comma).join(values),
            I
        ),
        LB,
        models.settings.csh.f_bracket("]")
    )

def format_tuple(v: tuple, exc: dict[int, str], i: str, lb: str) -> str:
    """Format a tuple value.
    
    """
    I = "" if models.settings.force_tuples_collapsed else i
    LB = "" if models.settings.force_tuples_collapsed else lb
    values = [concat(LB, format_value(V, exc, I, LB)) for V in v]
    return concat(
        models.settings.csh.f_bracket("("),
        indent(
            models.settings.csh.f_operator(models.settings.comma).join(values),
            I
        ),
        (models.settings.csh.f_operator(models.settings.comma)
         if len(values) == 1 else ""),
        LB,
        models.settings.csh.f_bracket(")")
    )

def format_dict(v: dict, exc: dict[int, str], i: str, lb: str) -> str:
    """Format a dict value.
    
    """
    I = "" if models.settings.force_dicts_collapsed else i
    LB = "" if models.settings.force_dicts_collapsed else lb
    values = [concat(LB, format_value(k, exc, I, LB),
              models.settings.csh.f_operator(models.settings.colon),
              format_value(V, exc, I, LB)) for k, V in v.items()]
    return concat(
        models.settings.csh.f_bracket("{"),
        indent(
            models.settings.csh.f_operator(models.settings.comma).join(values),
            I),
        LB,
        models.settings.csh.f_bracket("}")
    )

def format_class(v: object) -> str:
    """Format a class (not instance) value.
    
    """
    return models.settings.csh.f_operator(".").join([
        models.settings.csh.f_class(class_name) for class_name
        in v.__qualname__.split(".")])

def format_func(v, lb: str) -> str:
    """Format a function value.
    
    """
    if "__wrapped__" in dir(v):
        wrapped = True
    else:
        wrapped = False
    quals: list[str] = v.__qualname__.split(".")
    quals_formatted = [models.settings.csh.f_class(q) for q in quals[:-1]] + [
        models.settings.csh.f_function(quals[-1])]
    QUAL = models.settings.csh.f_operator(".").join(quals_formatted)
    if wrapped and "\n" in lb:
        return QUAL + models.settings.csh.f_comment(models.settings.comment +
            "wrapped")
    return QUAL

def format_enum(v: enum.Enum) -> str:
    """Format a `enum.Enum` value.
    
    """
    values = str(v).split(".")
    return concat(
        models.settings.csh.f_operator(".").join([
            models.settings.csh.f_class(class_name)
            for class_name in values[:-1]]),
        models.settings.csh.f_operator("."),
        models.settings.csh.f_enum(values[-1]))


def format_prepr(v: types.prepr, exc: dict[int, str], i: str, lb: str) -> str:
    """Format a `prepr` value.
    
    """
    if id(v._inst) in exc:
        return models.settings.csh.f_error(v._variable)
    I = "" if models.settings.force_sub_preprs_collapsed else i
    LB = "" if models.settings.force_sub_preprs_collapsed else lb
    return v._build_simple(exc, I, LB)


def format_value(v, exc: dict[int, str], i: str, lb: str) -> str:
    """Format a given text value with the global colorspace depending on its
    type.
    
    """
    TYPE = type(v)
    if TYPE == str:
        return format_str(v)
    if TYPE in [int, float]:
        return format_num(v)
    if TYPE == bool or v is None:
        return format_bool_none(v)
    if TYPE == list:
        return format_list(v, exc, i, lb)
    if TYPE == tuple:
        return format_tuple(v, exc, i, lb)
    if TYPE == dict:
        return format_dict(v, exc, i, lb)
    if inspect.isclass(v):
        return format_class(v)
    if TYPE in [builtin_types.FunctionType, builtin_types.BuiltinFunctionType,
                builtin_types.BuiltinMethodType]:
        return format_func(v, lb)
    if isinstance(v, enum.Enum):
        return format_enum(v)
    if isinstance(v, types.pstr):
        return format_prepr(v._prepr, exc, i, lb)
    if hasattr(v, "__repr__"):
        try:
            _prepr = v.__repr__(return_prepr=True)
        except Exception:
            _prepr = None
        if isinstance(_prepr, types.prepr):
            return format_prepr(_prepr, exc, i, lb)
    if hasattr(v, "__str__"):
        try:
            _prepr = v.__str__(return_prepr=True)
        except Exception:
            _prepr = None
        if isinstance(_prepr, types.prepr):
            return format_prepr(_prepr, exc, i, lb)
    return models.settings.csh.f_other(attempt_str(v))



from . import types
import dataclasses


class CSHandler:
    """Used to create a colorspace handler given a valid `types.Colorspace` (e.g. those
    found in `Colorspace`).
    
    """
    def __init__(self, cs: types.Colorspace) -> None:
        self.cs = cs
    def f_function(self, __text) -> str:
        return self.cs.c_function + __text + self.cs.c_reset
    def f_class(self, __text) -> str:
        return self.cs.c_class + __text + self.cs.c_reset
    def f_string(self, __text) -> str:
        return self.cs.c_string + __text + self.cs.c_reset
    def f_number(self, __text) -> str:
        return self.cs.c_number + __text + self.cs.c_reset
    def f_other(self, __text) -> str:
        return self.cs.c_other + __text + self.cs.c_reset
    def f_variable(self, __text) -> str:
        return self.cs.c_variable + __text + self.cs.c_reset
    def f_attribute(self, __text) -> str:
        return self.cs.c_attribute + __text + self.cs.c_reset
    def f_argument(self, __text) -> str:
        return self.cs.c_argument + __text + self.cs.c_reset
    def f_operator(self, __text) -> str:
        return self.cs.c_operator + __text + self.cs.c_reset
    def f_comment(self, __text) -> str:
        return self.cs.c_comment + __text + self.cs.c_reset
    def f_bracket(self, __text) -> str:
        return self.cs.c_bracket + __text + self.cs.c_reset
    def f_error(self, __text) -> str:
        return self.cs.c_error + __text + self.cs.c_reset
    def f_boolean(self, __text) -> str:
        return self.cs.c_boolean + __text + self.cs.c_reset
    def f_enum(self, __text) -> str:
        return self.cs.c_enum + __text + self.cs.c_reset
    def f_reset(self, __text) -> str:
        return self.cs.c_reset + __text + self.cs.c_reset


class Colorspace:

    class rgbfull:
        """The most accurate of the colorspaces that uses full RGB.
        
        """
        c_function = "\033[38;2;219;219;170m"
        c_class = "\033[38;2;77;200;176m"
        c_string = "\033[38;2;205;144;120m"
        c_number = "\033[38;2;180;205;168m"
        c_other = "\033[38;2;211;211;211m"
        c_variable = "\033[38;2;156;219;253m"
        c_attribute = "\033[38;2;156;219;253m"
        c_argument = "\033[38;2;156;219;253m"
        c_operator = "\033[38;2;211;211;211m"
        c_comment = "\033[38;2;109;109;109m"
        c_bracket = "\033[38;2;254;214;24m"
        c_error = "\033[38;2;243;71;70m"
        c_boolean = "\033[38;2;86;156;213m"
        c_enum = "\033[38;2;80;193;253m"
        c_reset = "\033[0m"
    class rgb256:
        """Less accurate than `rgbfull` but has slightly greater compatability in other
        terminals.
        
        """
        c_function = "\033[38;5;179m"
        c_class = "\033[38;5;43m"
        c_string = "\033[38;5;180m"
        c_number = "\033[38;5;151m"
        c_other = "\033[38;5;251m"
        c_variable = "\033[38;5;153m"
        c_attribute = "\033[38;5;153m"
        c_argument = "\033[38;5;153m"
        c_operator = "\033[38;5;251m"
        c_comment = "\033[38;5;241m"
        c_bracket = "\033[38;5;221m"
        c_error = "\033[38;5;203m"
        c_boolean = "\033[38;5;26m"
        c_enum = "\033[38;5;26m"
        c_reset = "\033[0m"
    class rgb8:
        """The least accurate but most compatible colorspace. Works in almost all
        terminals.
        
        """
        c_function = "\033[33m"
        c_class = "\033[32m"
        c_string = "\033[31m"
        c_number = "\033[36m"
        c_other = "\033[37m"
        c_variable = "\033[36m"
        c_attribute = "\033[36m"
        c_argument = "\033[36m"
        c_operator = "\033[37m"
        c_comment = "\033[30m"
        c_bracket = "\033[33m"
        c_error = "\033[31m"
        c_boolean = "\033[34m"
        c_enum = "\033[34m"
        c_reset = "\033[0m"
    class none:
        """A colorspace that will apply no terminal formatting.
        
        """
        c_function = ""
        c_class = ""
        c_string = ""
        c_number = ""
        c_other = ""
        c_variable = ""
        c_attribute = ""
        c_argument = ""
        c_operator = ""
        c_comment = ""
        c_bracket = ""
        c_error = ""
        c_boolean = ""
        c_enum = ""
        c_reset = ""
    @dataclasses.dataclass
    class custom:
        """A custom colorspace.
        
        """
        c_function: str
        c_class: str
        c_string: str
        c_number: str
        c_other: str
        c_variable: str
        c_attribute: str
        c_argument: str
        c_operator: str
        c_comment: str
        c_bracket: str
        c_error: str
        c_boolean: str
        c_enum: str
        c_reset: str


class settings:
    """Global settings used by all `prepr` instances. `csh` (colorspace handler) must be
    created with the `CSHandler` class and a valid `Colorspace`, e.g.
    `CSHandler(Colorspace.rgb256)`.
    
    """
    indent: str = "    "
    line_break: str = "\n"
    comma: str = ", "
    colon: str = ": "
    semicolon: str = "; "
    equals: str = " = "
    comment: str = " # "
    csh = CSHandler(Colorspace.rgbfull)
    force_lists_collapsed: bool = False
    force_tuples_collapsed: bool = False
    force_dicts_collapsed: bool = False
    force_sub_preprs_collapsed: bool = False
    @staticmethod
    def default() -> None:
        settings.indent = "    "
        settings.line_break = "\n"
        settings.comma = ", "
        settings.colon = ": "
        settings.semicolon = "; "
        settings.equals = " = "
        settings.comment = " # "
        settings.force_lists_collapsed = False
        settings.force_tuples_collapsed = False
        settings.force_dicts_collapsed = False
        settings.force_sub_preprs_collapsed= False
    @staticmethod
    def minimal() -> None:
        settings.indent = ""
        settings.line_break = ""
        settings.comma = ","
        settings.colon = ":"
        settings.semicolon = ";"
        settings.equals = "="
        settings.comment = "#"
        settings.force_lists_collapsed = True
        settings.force_tuples_collapsed = True
        settings.force_dicts_collapsed = True
    def update(
        self,
        indent: str = types.MISSING,
        line_break: str = types.MISSING,
        comma: str = types.MISSING,
        colon: str = types.MISSING,
        semicolon: str = types.MISSING,
        equals: str = types.MISSING,
        comment: str = types.MISSING,
        csh: CSHandler = types.MISSING,
        force_lists_collapsed: bool = types.MISSING,
        force_tuples_collapsed: bool = types.MISSING,
        force_dicts_collapsed: bool = types.MISSING,
        force_sub_preprs_collapsed: bool = types.MISSING
    ):
        """Batch-update settings.
        
        """
        pairs = [
            ("indent", indent),
            ("line_break", line_break),
            ("comma", comma),
            ("colon", colon),
            ("semicolon", semicolon),
            ("equals", equals),
            ("comment", comment),
            ("csh", csh),
            ("force_lists_collapsed", force_lists_collapsed),
            ("force_tuples_collapsed", force_tuples_collapsed),
            ("force_dicts_collapsed", force_dicts_collapsed),
            ("force_sub_preprs_collapsed", force_sub_preprs_collapsed)
        ]
        for name, value in pairs:
            if value != types.MISSING:
                setattr(self, name, value)


class pstr(types.pstr):
    """A `str` instance with an added `_prepr` attribute to store the `prepr`
    instance used to create it. Used for nested reprs.
    
    """
    def __new__(self, text, *args, **kwargs):
        return str.__new__(self, text)
    def __init__(self, text: str, _prepr: types.prepr):
        str.__init__(text)
        self._prepr = _prepr



"""A system for creating pretty class representations.

:copyright: (c) 2022 Tanner B. Corcoran
:license: MIT, see LICENSE for more details.
"""


__title__ = "prepr"
__author__ = "Tanner B. Corcoran"
__email__ = "tannerbcorcoran@gmail.com"
__license__ = "MIT License"
__copyright__ = "Copyright (c) 2022 Tanner B. Corcoran"
__version__ = "0.0.5"
__description__ = "A system for creating pretty class representations"
__url__ = "https://github.com/tanrbobanr/prepr"
__download_url__ = "https://pypi.org/project/prepr"


from .types import pstr
from .models import CSHandler, Colorspace, settings
from .main import prepr
